
static char rcsid[] =
	"$Id: pvm_test.c,v 1.37 1999/12/13 18:37:18 pvmsrc Exp $";

#include "pvm_test.h"
#include "globals.h"

extern void term(); 					/* Interruption handler */

void dump_tests();
void read_args();
void do_test();
void usage();


int
main( argc, argv )
int	argc;
char **argv;
{
	char	*myversion;
	char	*myarch;

	int		any;					/* were any tests run at all? */
	int		failed;					/* # of tests that failed */
	int		info;					/* dummy variable for return vals */
	int		iter;					/* iteration # for tests */
	int		mytid; 					/* PVM tid */
	int		num_diff_arch; 			/* # of different data formats */
	int		perform;				/* flag to check to do tests */
	int		shuffle[ MAX_HOST ]; 	/* shuffled hostpool index array */
	int		test_num; 				/* current test module processed */
	int		tmp;


	/**********************************/
	/*  Miscelleanous Initialization  */
	/**********************************/

	/*****  Set stdout and stderr *****/

	setbuf( stdout, NULL );
	dup2( fileno( stdout ), fileno( stderr ) );

	/*****  Get PVM_ARCH *****/

	if ( (myarch = getenv( "PVM_ARCH" )) == NULL )
	{
		fprintf( stderr, 
				"\tThe PVM_ARCH environment variable should be set\n" );
		exit( 1 );
	}

	assert( (hostarch[0] = strdup( myarch )) != NULL );

	/***** Get name of this host *****/

	assert( hostpool[0] = (char *) malloc( sizeof( char )
			* MAXHOSTNAMELEN ) );

	if ( gethostname( hostpool[0], MAXHOSTNAMELEN ) == -1 )
	{
		fprintf( stderr, "\tgethostname() call failed" );
		exit( 1 );
	}

	num_hosts = 1;

	/***** Get version of PVM on this host *****/

	if ( pvm_version() == NULL )
	{
		fprintf( stderr, "\tpvm_version(): Failed\n" );
		exit( 1 );
	}

	assert( (myversion = strdup( pvm_version() )) != NULL );

	/***** Setting the default parameters *****/

	messages_length		=	MESSAGES_LENGTH_DEFAULT;
	max_size			=	MAX_SIZE_DEFAULT;
	incr				=	INCR_DEFAULT;
	start_size			=	START_SIZE_DEFAULT;
	num_messages		=	NUM_MESSAGES_DEFAULT;
	all_combos			=	ALL_COMBOS_DEFAULT;
	timing_average		=	TIMING_AVERAGE_DEFAULT;
	timing_total		=	TIMING_TOTAL_DEFAULT;
	all_tests			=	ALL_TESTS_DEFAULT;
	all_messaging		=	ALL_MESSAGING_DEFAULT;
	all_routine			=	ALL_ROUTINE_DEFAULT;
	all_funnel_nodata	=	ALL_FUNNEL_NODATA_DEFAULT;
	all_funnel_data		=	ALL_FUNNEL_DATA_DEFAULT;
	all_head_nodata		=	ALL_HEAD_NODATA_DEFAULT;
	all_head_data		=	ALL_HEAD_DATA_DEFAULT;
	all_triangle_nodata	=	ALL_TRIANGLE_NODATA_DEFAULT;
	all_triangle_data	=	ALL_TRIANGLE_DATA_DEFAULT;
	all_perf			=	ALL_PERF_DEFAULT;

	/***** Parse the arguments *****/

	read_args( argc, argv );

	/***** Input file parsing *****/

	if ( infilename != NULL )
		input_parse();
	
	else
		outfilename = copy_str( "stdout" );

	if ( outfile == NULL )
		outfile = stdout;

	setbuf( outfile, NULL );

	/***** Set Trecv Timeval *****/

	TRECV_TM.tv_sec = SEC_TRECV;
	TRECV_TM.tv_usec = 0;

	/***** Welcome message *****/

	fprintf( outfile, "------------------------------------\n" );
	fprintf( outfile, "PVM tester --- version 1.2\n" );
	fprintf( outfile, "------------------------------------\n" );

	/***** Check parameters *****/

	if ( start_size > max_size )
	{
		fprintf( stdout,
				"\tError: start size > max size in input file\n" );
		exit( 1 );
	}

	/***** Update Tests Setting *****/

	perform = 0;

	if ( all_tests )			perform |= ALL_TYPES;
	if ( all_routine )			perform |= TYPE_ROUTINE;
	if ( all_messaging )		perform |= TYPE_MESSAGING;
	if ( all_funnel_nodata )	perform |= TYPE_FUNNEL_NODATA;
	if ( all_funnel_data )		perform |= TYPE_FUNNEL_DATA;
	if ( all_head_nodata )		perform |= TYPE_HEAD_NODATA;
	if ( all_head_data )		perform |= TYPE_HEAD_DATA;
	if ( all_triangle_nodata )	perform |= TYPE_TRIANGLE_NODATA;
	if ( all_triangle_data )	perform |= TYPE_TRIANGLE_DATA;
	if ( all_perf )				perform |= TYPE_PERF;

	/***** Set handlers *****/

	signal( SIGINT, term );
	signal( SIGTERM, term );

	/***** Check Starting of Virtual Machine *****/

	info = pvm_setopt( PvmAutoErr, 0 );
	CKERR_FP( outfile, info, "pvm_setopt", NOHOST, NOHANDLER );

	tmp = pvm_start_pvmd( 0, NULL, 1 );

	if ( tmp != PvmDupHost )
	{
		CKERR_FP( outfile, tmp, "pvm_start_pvmd", NOHOST,
			{ pvm_perror( "PVM Tester" ); exit( 1 ); } );

		fprintf( outfile, "PVM started on host %s.\n", hostpool[0] );

		newpvm++;
	}

	else
		fprintf( outfile, "PVM already running.\n" );

#ifndef DEBUG

	/***** Leave PVM error messages disabled *****/

	if ( info < 0 )
	{
		fprintf( outfile,
				"\tWarning:  Extra diagnostic messages may appear.\n" );
	}

#else

	/* Reset PVM error messages */

	info = pvm_setopt( PvmAutoErr, info );

	CKERR_FP( outfile, info, "pvm_setopt", NOHOST, NOHANDLER );

	/* Catch Slave Errors */

	info = pvm_catchout( outfile );

	CKERR_FP( outfile, info, "pvm_catchout", NOHOST, NOHANDLER );

#endif

	/***** Get my tid *****/

	mytid = pvm_mytid();

	CKERR_FP( outfile, mytid, "pvm_mytid", NOHOST, term( HIGH_LEVEL ) );

	/***** Remove all tasks *****/

	if ( remove_all_tasks() == FAILED )
		term( HIGH_LEVEL );

	/***** Remove all hosts *****/

	if ( remove_all_hosts() == FAILED )
		term( HIGH_LEVEL );

	/***** Check hosts *****/

	if ( check_hosts() == FAILED )
		term( HIGH_LEVEL );

	/***** Check Version *****/

	info = check_version( myversion );

	if ( info == FAILED )
		term( HIGH_LEVEL );

	/***** Get all arch *****/

	if ( get_all_arch() == FAILED )
		term( HIGH_LEVEL );

	/***** Get number of different arch *****/

	if ( (num_diff_arch = get_num_diff_arch()) < 0 )
	{
		fprintf( outfile,
			"\tUnable to compute the number of different arches\n" );
		term( HIGH_LEVEL );
	}

	/***** Compute the shuffle *****/

	if ( find_shuffle( num_diff_arch, shuffle ) < 0 )
	{
		fprintf( outfile, "\tUnable to compute the shuffle\n" );
		term( HIGH_LEVEL );
	}

	/***** Remove all hosts *****/

	if ( remove_all_tasks() == FAILED )
		term( HIGH_LEVEL );

	/***** Remove all hosts *****/

	if ( remove_all_hosts() == FAILED )
		term( HIGH_LEVEL );

	/***** Start Group Server *****/

	if ( start_group_server() == FAILED )
		term ( HIGH_LEVEL );

	/***** Printout Configuration *****/

	print_out_misc( myversion );

	/***************/
	/*  Main Loop  */
	/***************/

	any = 0;

	failed = 0;

	for ( test_num=0 ; test_num < NUMBER_OF_TESTS ; test_num++ )
	{
		for ( iter=0 ; iter < num_iters ; iter++ )
		{
			/* We perform this test */
	
			if ( ( ( module_list[ test_num ].type & perform )
				  && module_list[ test_num ].flag.perform != NOPERFORM )
				|| module_list[ test_num ].flag.perform == PERFORM )
			{
				do_test( test_num, iter, myarch, num_diff_arch, shuffle,
						&any, &failed );
			}
		}
	}

	if ( !any )
		fprintf( outfile, "\n***  No Tests Selected ***\n" );

	else
	{
		if ( failed )
			fprintf( outfile, "\n%d Tests Failed.\n", failed );
		
		else
			fprintf( outfile, "\nAll %d Tests Passed.\n", any );
	}

	term( OK_LEVEL );
}


void
do_test( test_num, iter, myarch, num_diff_arch, shuffle, any, failed )
int test_num;
int iter;
char *myarch;
int num_diff_arch;
int shuffle[ MAX_HOST ];
int *any;
int *failed;
{
	FILE	*slaveout;				/* file ptr for slave catchout */

	char	slaveoutfile[255];		/* file name for slave catchout */

	char	c;						/* dummy slave output char */

	int		i, j, k; 				/* loop indices */
	int		index; 					/* size of window in shuffle */
	int		info;
	int		num_wanted; 			/* # of hosts wanted after a test */
	int		num_wanted_ef; 			/* actual number of hosts wanted */
	int		req; 					/* # of hosts reqd current test */
	int		status; 				/* temporary status of a test */
	int		sub_shuffle_pool[ MAX_HOST ]; 		/* subset of shuffle */

	if ( num_iters > 1 )
	{
		fprintf( outfile, "test#%d(%d): %s  ",
				test_num + 1, iter + 1,
				module_list[ test_num ].description );
	}
	else
	{
		fprintf( outfile, "test#%d: %s  ",
				test_num + 1, module_list[ test_num ].description );
	}
	fflush( outfile );
	
	sprintf( slaveoutfile, "slaveout.%d", test_num + 1 );
	
	slaveout = fopen( slaveoutfile, "w" );
	
	if ( slaveout != NULL )
	{
		info = pvm_catchout( slaveout );
	
		CKERR_FP( outfile, info, "pvm_catchout", NOHOST,
				{ fclose( slaveout ); slaveout = (FILE *) NULL; } );
	}
	
	req = module_list[ test_num ].flag.num_req_hosts;
	num_wanted = module_list[ test_num ].flag.after_hosts;
	
	if ( req == ALL_HOSTS )
		req = num_hosts;
	if ( req == ALL_ARCHS )
		req = num_diff_arch + 1;
	if ( req > num_hosts )
		req = num_hosts;
	if ( num_wanted == ALL_HOSTS )
		num_wanted = num_hosts;
	if ( num_wanted == ALL_ARCHS )
		num_wanted = num_diff_arch+1;
	if ( num_wanted > num_hosts )
		num_wanted = num_hosts;
	
	if ( req > num_diff_arch || req == 1 || req == num_hosts
			|| all_combos == 0 )
	{
		/****** No shuffling Problems ******/
	
		if ( allocate( shuffle, req, &num_wanted ) < 0 )
		{
			/* The allocation didn't succeed */
	
			module_list[ test_num ].status = NOT_COMPLETED;
	
			errprintf( "\tAllocation failed for some hosts\n" );
	
			if ( test_cleanup( test_num ) < 0 )
				term( HIGH_LEVEL );
		}
	
		/* We do the test */
	
		else
		{
			status = module_list[ test_num ].proc_master(
					test_num, shuffle, req,
					module_list[ test_num ].proc_args );
	
			if ( status == NOT_POSSIBLE )
			{
				module_list[ test_num ].status = PASSED;
	
				errprintf( "\tNot possible for arch combo:\n" );
				errprintf( "\t%s | ", myarch );
				for ( k=1 ; k < req ; k++ )
					errprintf( " %s ", hostarch[shuffle[k]] );
				for ( k=req ;
						k < module_list[ test_num ].flag.num_req_hosts ;
						k++ )
				{
					errprintf( " %s ", hostarch[ shuffle[ k % req ] ] );
				}
				errprintf( "\n" );
			}
	
			module_list[test_num].status = status;
	
			if ( ( info = unexpected_host_killed( num_wanted - 1 ) )
					== FAILED )
			{
				errprintf( "\tNot the right # of hosts after test\n" );
	
				module_list[ test_num ].status = NOT_COMPLETED;
			}
	
			if ( info == -2 )
			{
				errprintf( "\tFailure of checking function\n" );
	
				module_list[ test_num ].status = NOT_COMPLETED;
			}
	
			if ( test_cleanup( test_num ) < 0 )
				term( HIGH_LEVEL );
		}
	}
	
	/***** We must try all the combinations *****/
	
	else
	{
		sub_shuffle_pool[0] = shuffle[0];
	
		module_list[ test_num ].status = PASSED;
	
		for ( i=0 ; i < num_diff_arch - req + 2 ; i++ )
		{
			sub_shuffle_pool[1] = shuffle[ i + 1 ];
	
			if ( req == 2 )
				index = 1;
			else
				index = num_diff_arch - req - i + 2;
	
			for ( j=0 ; j < index ; j++ )
			{
				num_wanted_ef = num_wanted;
	
				for( k=1 ; k < req - 1 ; k++ )
				{
					sub_shuffle_pool[ k + 1 ] =
							shuffle[ i + j + k + 1 ];
				}
	
				/* allocation didn't succeed */
	
				if ( allocate( sub_shuffle_pool, req, &num_wanted_ef )
						< 0 )
				{
					errprintf( "\tAllocation failed for some hosts\n" );
	
					module_list[ test_num ].status = NOT_COMPLETED;
	
					if ( remove_all_hosts() < 0 )
					{
						/* Cannot remove hosts */
						errprintf( "\tCannot remove hosts\n" );
						term( HIGH_LEVEL );
					}
				}
	
				/* We do the test */
	
				else
				{
					status = module_list[ test_num ].proc_master(
							test_num, sub_shuffle_pool, req,
							module_list[ test_num ].proc_args );
	
					if ( status == FAILED )
					{
						module_list[ test_num ].status = FAILED;
	
						errprintf( "\tFailed for arch combo: " );
						errprintf( "\t%s ", myarch );
						for ( k=1 ; k < req ; k++ )
						{
							errprintf( " %s ",
									hostarch[ sub_shuffle_pool[ k ] ] );
						}
						errprintf( "\n" );
					}
	
					if ( status == NOT_POSSIBLE )
					{
						module_list[ test_num ].status = NOT_POSSIBLE;
	
						errprintf( "\tNot possible for arch combo:" );
						errprintf( "%s | ", myarch );
						for ( k=1 ; k < req ; k++ )
						{
							errprintf( " %s ",
									hostarch[ sub_shuffle_pool[ k ] ] );
						}
						for ( k=req ; k <
							module_list[ test_num ].flag.num_req_hosts ;
							k++ )
						{
							errprintf( " %s ", myarch );
						}
						errprintf( "\n" );
					}
	
					if ( unexpected_host_killed( num_wanted_ef - 1 )
							< 0 )
					{
						errprintf( "\tSome hosts unexpectedly died\n" );
	
						module_list[ test_num ].status = FAILED;
					}
	
					if ( test_cleanup( test_num ) < 0 )
						term( HIGH_LEVEL );
				}
			}
		}
	}
	
	print_test_result( test_num );
	
	dump_errmsgs();
	
	(*any)++;
	
	if ( module_list[ test_num ].status != PASSED )
	{
		/* Dump any slave output */
	
		if ( slaveout != NULL )
		{
			fclose( slaveout );
	
			slaveout = fopen( slaveoutfile, "r" );
	
			if ( slaveout != NULL )
			{
				while( (c = getc( slaveout )) != (char) EOF )
					printf( "%c", c );
				
				fclose( slaveout );
	
				slaveout = (FILE *) NULL;
			}
	
			else
				printf( "\tError Dumping Slave Output\n" );
	
			unlink( slaveoutfile );
		}
	
		(*failed)++;
	}
	
	if ( slaveout != NULL )
	{
		fclose( slaveout );

		unlink( slaveoutfile );
	}
}


void
read_args( argc, argv )
int argc;
char **argv;
{
	FILE	*fp;

	void	*pattbuff;

	char	tmp[1024];

	char	*test_patt;

	int		i, j, k, m;
	int		do_usage;
	int		len;
	int		cc;

	slave_spawn_flags = PvmTaskDefault;

	infilename = (char *) NULL;

	num_iters = 1;

	vflag = 0;

	do_usage = 0;

	for ( i=1 ; i < argc ; i++ )
	{
		if ( argv[i][0] == '-' )
		{
			k = i + 1;

			len = strlen( argv[i] );

			for ( j=0 ; j < len ; j++ )
			{
				switch ( argv[i][j] )
				{
					case 'D':
						slave_spawn_flags |= PvmTaskDebug;
						break;

					case 'N':
						num_iters = atoi( argv[k++] );
						num_iters = ( num_iters > 0 ) ? num_iters : 1;
						break;
					case 'H':
					case 'h':
						usage();
						break;

					case 'V':
						dump_tests();
						break;

					case 'v': vflag++; break;

					case '-': break;

					default:
					{
						fprintf( stderr, "\nUnknown Option -%c\n",
								argv[i][j] );

						do_usage++;

						break;
					}
				}
			}

			i = k - 1;
		}

		else if ( argv[i][0] == '+' || argv[i][0] == '#' )
		{
#ifdef USE_GNU_REGEX

			sprintf( tmp, ".*%s.*", &(argv[i][1]) );

			test_patt = copy_str( tmp );

			pattbuff = pvmcompileregex( test_patt );

			if ( pattbuff )
			{
				for ( m=0 ; m < NUMBER_OF_TESTS ; m++ )
				{
					if ( pvmmatchregex( pattbuff,
							module_list[m].description ) )
					{
						module_list[m].flag.perform =
							( argv[i][0] == '+' ) ? PERFORM : NOPERFORM;
					}
				}

				pvmfreeregex( &pattbuff );
			}

			else
			{
				fprintf( stderr, "Error:  Invalid Regular Expression" );
				fprintf( stderr, " \"%s\"\n", test_patt );
			}

			free( test_patt );

#else

			sprintf( tmp, "*%s*", &(argv[i][1]) );

			test_patt = copy_str( tmp );

			for ( m=0 ; m < NUMBER_OF_TESTS ; m++ )
			{
				if ( pvmmatchstring( module_list[m].description,
						test_patt ) )
				{
					module_list[m].flag.perform =
						( argv[i][0] == '+' ) ? PERFORM : NOPERFORM;
				}
			}

			free( test_patt );

#endif
		}

		else
		{
			if ( (fp = fopen( argv[i], "r" )) != NULL )
			{
				fclose( fp );

				if ( infilename == NULL )
					infilename = copy_str( argv[i] );
			
				else
				{
					fprintf( stderr, "\nError on \"%s\":  ", argv[i] );
					fprintf( stderr,
							"Input File \"%s\" Already Specified.\n",
							infilename );

					do_usage++;
				}
			}

			else if ( parse_option( argv[i] ) <= 0 )
			{
				fprintf( stderr, "\nUnknown Option %s\n", argv[i] );

				do_usage++;
			}
		}
	}

	if ( do_usage )
		usage();
}


void
usage()
{
	printf( "\nusage:  pvm_test " );
	printf( "[ -VDHhv ] [ -N <niters> ] " );
	printf( "[ <input file> ] " );
	printf( "[ <+/#><regex> ] " );
	printf( "\n\t\t" );
	printf( "[ <option>=<value> ] " );
	printf( "\n" );

	printf( "\n" );

	printf( "where:\n" );
	printf( "------\n" );

	printf( "<input file>      =  " );
	printf( "(Optional) Test Configuration File\n" );

	printf( "<+/#><regex>      =  " );
	printf( "Select / Deselect Tests Matching " );
	printf( "Regular Expr \"<regex>\"\n" );

	printf( "<option>=<value>  =  " );
	printf( "Set Test Option \"<option>\" to \"<value>\"\n" );

	printf( "-V                =  Dump List of Test Descriptions\n" );

	printf( "-D                =  Spawn Test Slaves in Debug Mode\n" );

	printf( "-N <niters>       =  " );
	printf( "Number of Iterations for Each Test\n" );

	printf( "-H/h              =  Print This Help Information\n" );

	printf( "-v                =  Verbose Operation\n" );

	printf( "\n" );

	printf( "Regular Expression Examples:\n" );
	printf( "----------------------------\n" );

	printf( "\n" );

	printf( "1. Run pvm_send() tests, but not pvm_sendsig() tests.\n" );
	printf( "\n\tpvm_test +send #sendsig\n\n" );

	printf( "2. Run head-head tests with InPlace packing.\n" );
	printf( "\t(Note the '' around the regex to protect '*' " );
	printf( "from the shell.)\n" );
	printf( "\n\tpvm_test +'head-head*InPlace'\n\n" );

	printf( "Note:  By Default All Expressions Support Standard\n" );
	printf( "Unix-based '*' Globbing, Unless PVM is Compiled with\n" );
	printf( "the USE_GNU_REGEX flag where the Standard Regex\n" );
	printf( "Format (e.g. '.*') is Used.\n" );

	printf( "\n" );

	printf( "All Expressions are Automatically Enclosed in\n" );
	printf( "Wildcard Pairs ('*'/'*' or '.*'/'.*') for Searching\n" );
	printf( "Anywhere Within the Test Description Text.\n" );

	printf( "\n" );
	printf( "\n" );

	printf( "PVM Test Suite Options:\n" );
	printf( "-----------------------\n" );

	printf( "\n" );

	printf( "host=<host>                 " );
	printf( "Use \"<host>\" in test configuration.\n" );

	printf( "outfile=<fname>             " );
	printf( "Write test results to file \"<fname>\".\n" );

	printf( "all_tests=<1/0>             " );
	printf( "Run / Don't Run All Tests.\n" );

	printf( "all_routine=<1/0>           " );
	printf( "Run / Don't Run All PVM Routine Tests.\n" );

	printf( "all_messaging=<1/0>         " );
	printf( "Run / Don't Run All Messaging Tests.\n" );

	printf( "all_perf=<1/0>              " );
	printf( "Run / Don't Run All Perf Tests (-v for stats).\n" );

	printf( "all_funnel_nodata=<1/0>     " );
	printf( "Run / Don't Run NODATA Funnel Messaging Tests\n" );

	printf( "all_funnel_data=<1/0>       " );
	printf( "Run / Don't Run DATA Funnel Messaging Tests\n" );

	printf( "all_head_nodata=<1/0>       " );
	printf( "Run / Don't Run NODATA Head-Head Messaging Tests\n" );

	printf( "all_head_data=<1/0>         " );
	printf( "Run / Don't Run DATA Head-Head Messaging Tests\n" );

	printf( "all_triangle_nodata=<1/0>   " );
	printf( "Run / Don't Run NODATA Triangle Messaging Tests\n" );

	printf( "all_triangle_data=<1/0>     " );
	printf( "Run / Don't Run DATA Triangle Messaging Tests\n" );

	printf( "all_combinations=<1/0>      " );
	printf( "Run / Don't Run All Host Combinations for Tests.\n" );

	printf( "start_length=<value>        " );
	printf( "Set Starting Message Length for Perf Tests.\n" );

	printf( "max_length=<value>          " );
	printf( "Set Maximum Message Length for Perf Tests.\n" );

	printf( "incr=<value>                " );
	printf( "Set Message Length Increment for Perf Tests.\n" );

	printf( "num_messages=<value>        " );
	printf( "Set # of Messages for Messaging Tests.\n" );

	printf( "messages_length=<value>     " );
	printf( "Set Message Length for Messaging Tests.\n" );

	printf( "funneling_proc=<value>      " );
	printf( "Set # of Hosts for Funneling Tests.\n" );

	printf( "total_timing=<1/0>          " );
	printf( "Do / Do Not Total Timings for Messaging Tests.\n" );

	printf( "average_timing=<1/0>        " );
	printf( "Do / Do Not Average Timings for Messaging Tests.\n" );

	printf( "\n" );

	printf( "Notes:\n" );
	printf( "    -> Tests denoted NODATA send empty messages.\n" );
	printf( "    -> Tests denoted DATA send messages packed " );
	printf( "with integers.\n" );
	printf( "    -> Message Lengths are in Integer Words.\n" );

	printf( "\n" );

	exit( 0 );
}


void
dump_tests()
{
	int		i;

	printf( "\n%d Tests:\n\n", NUMBER_OF_TESTS );

	for ( i=0 ; i < NUMBER_OF_TESTS ; i++ )
	{
		printf( "\ttest#%d:  %s\n",
				i + 1, module_list[i].description );
	}

	exit( 0 );
}

