
/* $Id: pvm_test.h,v 1.33 2001/02/05 22:57:24 pvmsrc Exp $ */


/**************/
/*  Includes  */
/**************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <limits.h>
#include <netdb.h>
#include <assert.h>
#include <memory.h>
#include <errno.h>
#include <assert.h>
#include <sys/param.h>
#include "pvm3.h"

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


/*******************/
/*  Tests Modules  */
/*******************/

/* Test Status */

#define PASSED				1
#define NOT_YET_PROCESSED	0
#define FAILED				-1
#define NOT_COMPLETED		-2
#define NOT_POSSIBLE		-3

#define ALL_HOSTS			999			/* all hosts for test */
#define ALL_ARCHS			666			/* one host of each arch */

#define NOPERFORM			2
#define PERFORM				1


/******************************/
/*  Host Pool and descriptor  */
/******************************/

#define MAX_HOST			1024		/* Max # of hosts in config */

#define TIMEOUT_MESSAGE		60			/* # of 1 sec. retries */
#define TIMEOUT_ADD_DELETE	30			/* # of 1 sec. retries */
#define TIMEOUT_KILL		30			/* # of 1 sec. retries */
#define BUSY_LOOP_PERF		50000		/* # of iters in busy loop */


/************************/
/*  Special Test Codes  */
/************************/

#define VERSION_TEST		-1


/********************/
/*  INPUT / OUTPUT  */
/********************/

#define MAX_LINE_LENGTH		255			/* Max len input file line */

#define OUTFILE_MAX_LENGTH	255			/* Max len output file name */

#define COMBOS_PREFIX				"all_combinations"
#define HOST_PREFIX					"host"
#define OUTFILE_PREFIX				"outfile"

#define MAX_LENGTH_PREFIX			"max_length"
#define START_LENGTH_PREFIX			"start_length"
#define INCR_PREFIX					"incr"

#define LENGTH_PREFIX				"messages_length"
#define NUM_PREFIX					"num_messages"
#define FUNNELING_PREFIX			"funneling_proc"
#define TIMING_TOTAL_PREFIX			"total_timing"
#define TIMING_AVERAGE_PREFIX		"average_timing"

#define ALL_TESTS_PREFIX			"all_tests"
#define ALL_ROUTINE_PREFIX			"all_routine"
#define ALL_MESSAGING_PREFIX		"all_messaging"
#define ALL_FUNNEL_NODATA_PREFIX	"all_funnel_nodata"
#define ALL_FUNNEL_DATA_PREFIX		"all_funnel_data"
#define ALL_HEAD_NODATA_PREFIX		"all_head_nodata"
#define ALL_HEAD_DATA_PREFIX		"all_head_data"
#define ALL_TRIANGLE_NODATA_PREFIX	"all_triangle_nodata"
#define ALL_TRIANGLE_DATA_PREFIX	"all_triangle_data"
#define ALL_PERF_PREFIX				"all_perf"


/*******************/
/*  MESSAGES TAGS  */
/*******************/

#define ACKNOWLEDGE		9		/* acknowledging slave start */
#define VERSION_TAG		10		/* returning version number */
#define JOINED			11		/* acknowledging a group joining */
#define SIGNAL			12		/* acknowledging an interrupt signal */
#define BYE				13		/* telling to the slave to exit */
#define TEST_VALUE		14		/* value coherency */
#define TEST_CAST		15		/* broadcast test */
#define TEST_BARRIER	16		/* barrier test */
#define TEST_REDUCE		17		/* reduce test */
#define TEST_GATHER		18		/* gather test */
#define TEST_SCATTER	19		/* scatter test */
#define TEST_MCAST		20		/* mcast test */
#define LATENCY			21		/* latency measurement messages */
#define BANDWIDTH		22		/* bandwidth measurements messages */
#define HEAD_HEAD		23		/* head-head test */
#define TRIANGLE		24		/* triangle test */
#define FUNNELING		25		/* funneling */
#define NOTIFY			26		/* notify test */
#define CONTEXT			27		/* context test */
#define MHFTAG0			28		/* mhf tag #1 */
#define MHFTAG1			29		/* mhf tag #2 */
#define MHFTAG2			30		/* mhf tag #3 */
#define ENDLESS			99		/* for non-busy waiting slaves */


/*******************/
/*  MISCELLENAOUS  */
/*******************/

#define TEST_SLAVE		"pvm_test_slave"	/* Generic slave name */

#define GROUP_NAME		"Testgrp"	/* group name */

#define SNOOZ			1			/* max delay for pvm_config races */

#define HIGH_LEVEL		-1			/* Exit code for term() */
#define OK_LEVEL		-2			/* End code for term() */

#define SEC_TRECV		10			/* # of seconds timeout for trecv */
#define SEC_NRECV		10			/* # of seconds timeout for nrecv */

#define NOWAIT			0			/* ack slave wait flags */
#define WAIT			1

#define NOBLOCK			0			/* ack slave block flags */
#define BLOCK			1

#define SILENT			0			/* timeout_wait_message() output */
#define VERBOSE			1

#define PACK			0			/* default message packing */
#define PACKF			1			/* use pvm_packf() instead */

#define UNPACK			0			/* default message unpacking */
#define UNPACKF			1			/* use pvm_unpackf() instead */


/***********************************************/
/*  PARAMETERS FOR THE LATENCY-BANDWIDTH TEST  */
/***********************************************/

#define TRIALS				10		/* # trials avg latency/bw comput */

#define MAX_SIZE_DEFAULT	1000
#define START_SIZE_DEFAULT	100		/* Default params for perf tests */
#define INCR_DEFAULT		100

#define MAX_MESSAGE_LENGTH	100000	/* Max allowable message length */

#define ALL_COMBOS_DEFAULT	0		/* default param for all_combos */


/******************************************************/
/*  PARAMETERS FOR THE GENERAL MESSAGE-PASSING TESTS  */
/******************************************************/

#define EMPTY			0		/* Empty Messages */
#define NOT_EMPTY		1		/* Not empty messages */

#define SEND			0		/* Use of pvm_send() */
#define SENDF			1		/* Use of pvm_send(), pvm_packf() */
#define PSEND			2		/* Use of pvm_psend() */
#define MCAST			3		/* Use of pvm_mcast() */

#define RECV			0		/* Use of pvm_recv() */
#define RECVF			1		/* Use of pvm_recv(), pvm_unpackf() */
#define PRECV			2		/* Use of pvm_precv() */
#define NRECV			3		/* Use of pvm_nrecv() */
#define TRECV			4		/* Use of pvm_trecv() */

#define NUM_MESSAGES_DEFAULT		10
#define MESSAGES_LENGTH_DEFAULT		100
#define FUNNELING_DEFAULT			4
#define TIMING_AVERAGE_DEFAULT 		0
#define TIMING_TOTAL_DEFAULT		0


/*****************************************/
/*  PARAMETERS FOR THE SUBSETS OF TESTS  */
/*****************************************/

#define ALL_TESTS_DEFAULT			0
#define ALL_ROUTINE_DEFAULT			0
#define ALL_MESSAGING_DEFAULT		0
#define ALL_FUNNEL_NODATA_DEFAULT	0
#define ALL_FUNNEL_DATA_DEFAULT		0
#define ALL_HEAD_NODATA_DEFAULT		0
#define ALL_HEAD_DATA_DEFAULT		0
#define ALL_TRIANGLE_NODATA_DEFAULT	0
#define ALL_TRIANGLE_DATA_DEFAULT	0
#define ALL_PERF_DEFAULT			0

#define TYPE_ROUTINE				1
#define TYPE_MESSAGING				2
#define TYPE_FUNNEL_NODATA			4
#define TYPE_FUNNEL_DATA			8
#define TYPE_HEAD_NODATA			16
#define TYPE_HEAD_DATA				32
#define TYPE_TRIANGLE_NODATA		64
#define TYPE_TRIANGLE_DATA			128
#define TYPE_PERF					256

#define ALL_TYPES					(int) 0xffff


/************/
/*  Macros  */
/************/

extern char *pvm_errlist[];

#define CKERR( _info, _routine, _host, _handler ) \
	if ( (_info) < 0 ) \
	{ \
		errprintf( "\t%s():  Failed with error %d (%s)", \
				_routine, _info, pvm_errlist[ -(_info) ] ); \
		if ( (_host) != NULL ) \
			errprintf( " for host %s", _host ); \
		errprintf( "\n" ); \
		_handler; \
	}

#define CKERR_FP( _fp, _info, _routine, _host, _handler ) \
	if ( (_info) < 0 ) \
	{ \
		fprintf( _fp, "\t%s():  Failed with error %d (%s)", \
				_routine, _info, pvm_errlist[ -(_info) ] ); \
		if ( (_host) != NULL ) \
			fprintf( _fp, " for host %s", _host ); \
		fprintf( _fp, "\n" ); \
		fflush( _fp ); \
		_handler; \
	}

#define EXPECTED( _routine, _val1, _val2, _host, _handler ) \
	if ( (_val1) != (_val2) ) \
	{ \
		errprintf( "\t%s():  expected ", _routine ); \
		if ( (_val1) < 0 ) \
			errprintf( "%d (%s) ", _val1, \
				pvm_errlist[ ( (_val1) < 0 ) ? -(_val1) : (_val1) ] ); \
		else \
			errprintf( "%d ", _val1 ); \
		errprintf( "but got " ); \
		if ( (_val2) < 0 ) \
			errprintf( "%d (%s)", _val2, pvm_errlist[ -(_val2) ] ); \
		else \
			errprintf( "%d", _val2 ); \
		if ( (_host) != NULL ) \
			errprintf( " for host %s", _host ); \
		errprintf( "\n" ); \
		_handler; \
	}

#define NOHOST				(char *) NULL

#define NOHANDLER			{ }

#define DEFAULT_SLAVE_ARGS	(int *) NULL

#define TIME_ELAPSED \
		( t2.tv_sec - t1.tv_sec + ( t2.tv_usec - t1.tv_usec ) * 1.E-6 )


/*************************************/
/*  Global Structures and Variables  */
/*************************************/

typedef void (*vfp)();
typedef int (*ifp)();

typedef struct {
	int		perform;		/* Do we perform the test or not */
	int		num_req_hosts;	/* Number of hosts required */
	int		after_hosts;	/* Number of Hosts after completion */
} test_flag;

#define NUM_ARGS    8

typedef struct {
	char 		description[60];	/* Description of the test module */
	test_flag	flag;				/* Flags */
	int			status;				/* Return code of the test */
	int			type;				/* test type flag */
	ifp			proc_master;		/* Pointer to the master function */
	int			proc_args[NUM_ARGS];/* Arguments to master */
	vfp			proc_slave;			/* Pointer to the slave function */
	int			proc_slave_args[NUM_ARGS];	/* Arguments to slave */
} test_module;

struct errmsg_struct
{
	char errmsg[255];
	/* (test_errmsg *) */ struct errmsg_struct *next;
};

typedef struct errmsg_struct test_errmsg;

#ifdef USE_GNU_REGEX
void *pvmcompileregex __ProtoGlarp__ (( char * ));
int pvmmatchregex __ProtoGlarp__ (( void *, char * ));
void pvmfreeregex __ProtoGlarp__ (( void ** ));
#else
int pvmmatchstring __ProtoGlarp__ (( char *, char * ));
#endif

#ifdef __STDC__
int errprintf __ProtoGlarp__(( const char *, ... ));
#else
int errprintf __ProtoGlarp__(( va_alist ));
#endif

int spawn_slave __ProtoGlarp__(( int, int *, int, char *, int, int * ));
int master_send __ProtoGlarp__(( int *, int, int, int, int,
		void *, int, int, char * ));
int master_recv __ProtoGlarp__(( int, int, int, void *, int, int,
		int *, char * ));
int master_pack_by_type __ProtoGlarp__(( int,
		void *, int, int, int, int, char * ));
int master_unpack_by_type __ProtoGlarp__(( char *, int, int, int, int,
		char * ));

void slave_send __ProtoGlarp__(( int *, int, int, int, int,
		void *, int, int ));
void slave_recv __ProtoGlarp__(( int, int, int, void *, int, int ));
void slave_recv_block __ProtoGlarp__(( void ));
void slave_pack_by_type __ProtoGlarp__(( int,
		void *, int, int, int, int ));
void slave_unpack_by_type __ProtoGlarp__(( char *, int, int, int,
		int ));

char *test_str();
char *copy_str();

void dump_errmsgs();

void funneling_slave();
void head_head_slave();
void triangle_slave();

int funneling();
int head_head();
int triangle();

int look_for_new_arch();
int parse_flag();
int parse_test_number();
int timeout_wait_kill();
int timeout_wait_message();
int wait_number_hosts();

