
static char rcsid[] =
	"$Id: test_util.c,v 1.37 1999/12/13 18:37:22 pvmsrc Exp $";

/*****************************/
/*  test_util.c              */
/*  miscelleanous functions  */
/*****************************/

#include "pvm_test.h"
#include "globals.h"

char *parse_test_pattern();


/***************/
/*  Terminate  */
/***************/

void
term( code )
int		code;
{
	dump_errmsgs();

	if ( code == HIGH_LEVEL )
	{
		fprintf( outfile, "\n\n" );
		fprintf( outfile, "********************\n" );
		fprintf( outfile, "* PVM Tester Exits *\n" );
		fprintf( outfile, "********************\n\n" );
	}

	else if ( code == OK_LEVEL )
	{
		fprintf( outfile, "\n\n" );
		fprintf( outfile, "--------------------\n" );
		fprintf( outfile, "| PVM Tester End |\n" );
		fprintf( outfile, "--------------------\n\n" );
	}

	else
		exit( code );

	if ( outfile != stdout )
		fclose( outfile );

	remove_all_tasks();
	remove_all_hosts();

	dump_errmsgs();

	( newpvm ) ? pvm_halt() : pvm_exit();

	exit( ( code == OK_LEVEL ) ? 0 : 1 );
}


/************************/
/*  Input file Parsing  */
/************************/

int
input_parse()
{
	FILE	*infile;

	char	line[ MAX_LINE_LENGTH ];

	char	*lineptr;

	int		linecount = 0;
	int		cc;

	/* File Open */

	if ( (infile = fopen( infilename, "r" )) == NULL )
	{
		fprintf( stderr, "Unable to open %s for input: ", infilename );
		perror( NULL );
		exit( 1 );
	}

	/* File parse */

	while ( fgets( line, MAX_LINE_LENGTH, infile ) != NULL )
	{
		linecount++;

		/* Skip any preceding white space */

		lineptr = line + strspn( line, " \t" );

		/* We ignore the comments */

		if ( *lineptr == '#' || *lineptr == '\n' )
			continue;

		else
		{
			cc = parse_option( lineptr );

			if ( cc < 0 )
			{
				fprintf( stderr, "Error in input file at line %d\n",
						linecount );

				exit( 1 );
			}

			else if ( !cc )
			{
				fprintf( stderr,
						"Error: unrecognized line %d in input file.\n",
						linecount );

				exit( 1 );
			}
		}
	}

	fclose( infile );

	return( 0 );
}


int
parse_option( str )
char *str;
{
	static int myhost_already_seen = 0;
	static int offset = 1;

	void	*pattbuff;

	char	*test_patt;

	int		value;
	int		flag;
	int		ret;
	int		fc;
	int		lc;
	int		i;

	ret = 0;

	if ( strncmp( str, HOST_PREFIX, strlen( HOST_PREFIX ) ) == 0 )
	{
		fc = strspn( str + strlen( HOST_PREFIX ), " \t:=" )
				+ strlen( HOST_PREFIX );

		lc = strcspn( str + fc, " \t\n" );

		*( str + fc + lc ) = (char) NULL;

		if ( strlen( str + fc ) > MAXHOSTNAMELEN )
		{
			fprintf( stderr, "Error: host name %s too long.\n",
					str + fc );
			return( FAILED );
		}

		if ( strcmp( hostpool[0], str + fc ) != 0 )
		{
			for ( i=1, flag=0 ; i < offset && !flag ; i++ )
			{
				if ( strcmp( str + fc, hostpool[i] ) == 0 )
				{
					fprintf( stderr, "Error: duplicate host %s.\n",
							str + fc );
					flag++;
				}
			}

			/***** Add a host to hostpool *****/

			if ( !flag )
			{
				hostpool[ offset ] = copy_str( str + fc );
#ifdef DEBUG
				fprintf( stderr, "input_parse(): Found host %s\n",
						hostpool[ offset ] );
#endif
				offset++;

				num_hosts++;
			}
		}

		/* it is myhost */

		else
		{
			if ( myhost_already_seen )
			{
				fprintf( stderr, "Error: duplicate host name %s\n",
						str + fc );
			}

			else
			{
#ifdef DEBUG
				fprintf( stderr, "input_parse(): Found local host %s\n",
						hostpool[0] );
#endif
				myhost_already_seen = 1;
			}
		}

		ret = PASSED;
	}

	/***** Do we execute this test *****/

	else if ( *str == '[' )
	{
		if ( (flag = parse_flag( str )) == FAILED )
			return( 0 );

		if ( flag )
		{
			if ( (test_patt = parse_test_pattern( str )) == NULL )
				return( FAILED );

#ifdef USE_GNU_REGEX

			pattbuff = pvmcompileregex( test_patt );

			if ( pattbuff )
			{
				for ( i=0 ; i < NUMBER_OF_TESTS ; i++ )
				{
					if ( pvmmatchregex( pattbuff,
							module_list[i].description ) )
						module_list[i].flag.perform = flag;
				}

				pvmfreeregex( &pattbuff );
			}

			else
			{
				fprintf( stderr, "Error:  Invalid Regular Expression" );
				fprintf( stderr, " \"%s\"\n", test_patt );
			}

			free( test_patt );

#else

			for ( i=0 ; i < NUMBER_OF_TESTS ; i++ )
			{
				if ( pvmmatchstring( module_list[i].description,
						test_patt ) )
					module_list[i].flag.perform = flag;
			}

#endif
		}

		ret = PASSED;
	}

	/***** Set outfile *****/

	else if ( strncmp( str,
			OUTFILE_PREFIX, strlen( OUTFILE_PREFIX ) ) == 0 )
	{
		fc = strspn( str + strlen( OUTFILE_PREFIX ), " \t:=" )
				+ strlen( OUTFILE_PREFIX );

		lc = strcspn( str + fc, " \t\n" );

		*( str + fc + lc ) = (char) NULL;

		if ( strlen( str + fc ) > OUTFILE_MAX_LENGTH )
		{
			fprintf( stderr, "Outfile too long\n" );
			return( FAILED );
		}

		outfilename = copy_str( str + fc );

		if ( !strcmp( outfilename, "stdout" ) )
			outfile = stdout;

		else if ( (outfile = fopen( outfilename, "w" )) == NULL )
		{
			fprintf( stderr, "fopen(): Unable to open %s\n",
					outfilename );
			return( FAILED );
		}

		ret = PASSED;
	}

	/***** Set max_size *****/

	else if ( ret = grok_line( str,
			MAX_LENGTH_PREFIX, &max_size,
			-1, MAX_MESSAGE_LENGTH, "Max Message Size" ) )
	{ }

	/***** Set start_size *****/

	else if ( ret = grok_line( str,
			START_LENGTH_PREFIX, &start_size,
			-1, MAX_MESSAGE_LENGTH, "Starting Message Size" ) )
	{ }

	/***** Set incr *****/

	else if ( ret = grok_line( str,
			INCR_PREFIX, &incr,
			-1, -1, "Message Size Increment" ) )
	{ }

	/***** Set num_message *****/

	else if ( ret = grok_line( str,
			NUM_PREFIX, &num_messages,
			-1, -1, "Number of Messages" ) )
	{ }

	/***** Set message_length *****/

	else if ( ret = grok_line( str,
			LENGTH_PREFIX, &messages_length,
			-1, -1, "Message Length" ) )
	{ }

	/***** set funneling *****/

	else if ( ret = grok_line( str,
			FUNNELING_PREFIX, &value,
			-1, -1, "Funneling Value" ) )
	{
		for ( i=0 ; i < NUMBER_OF_TESTS ; i++ )
		{
			if ( (module_list[i].type & TYPE_FUNNEL_NODATA)
					|| (module_list[i].type & TYPE_FUNNEL_DATA) )
			{
				module_list[i].flag.num_req_hosts = value;
				module_list[i].flag.after_hosts = value;
			}
		}
	}

	/***** set all_combos *****/

	else if ( ret = grok_line( str,
			COMBOS_PREFIX, &all_combos,
			0, 1, "Set All Combos" ) )
	{ }

	/***** set all_tests *****/

	else if ( ret = grok_line( str,
			ALL_TESTS_PREFIX, &all_tests,
			0, 1, "Set All Tests" ) )
	{ }

	/***** set all_messaging *****/

	else if ( ret = grok_line( str,
			ALL_MESSAGING_PREFIX, &all_messaging,
			0, 1, "Set All Messaging" ) )
	{ }

	/***** set all_routine *****/

	else if ( ret = grok_line( str,
			ALL_ROUTINE_PREFIX, &all_routine,
			0, 1, "Set All Routine" ) )
	{ }

	/***** set all_funnel_nodata *****/

	else if ( ret = grok_line( str,
			ALL_FUNNEL_NODATA_PREFIX, &all_funnel_nodata,
			0, 1, "Set All Funnel No Data" ) )
	{ }

	/***** set all_funnel_data *****/

	else if ( ret = grok_line( str,
			ALL_FUNNEL_DATA_PREFIX, &all_funnel_data,
			0, 1, "Set All Funnel Data" ) )
	{ }

	/***** set all_head_nodata *****/

	else if ( ret = grok_line( str,
			ALL_HEAD_NODATA_PREFIX, &all_head_nodata,
			0, 1, "Set All Head to Head No Data" ) )
	{ }

	/***** set all_head_data *****/

	else if ( ret = grok_line( str,
			ALL_HEAD_DATA_PREFIX, &all_head_data,
			0, 1, "Set All Head to Head Data" ) )
	{ }

	/***** set all_triangle_nodata *****/

	else if ( ret = grok_line( str,
			ALL_TRIANGLE_NODATA_PREFIX, &all_triangle_nodata,
			0, 1, "Set All Triangle No Data" ) )
	{ }

	/***** set all_triangle_data *****/

	else if ( ret = grok_line( str,
			ALL_TRIANGLE_DATA_PREFIX, &all_triangle_data,
			0, 1, "Set All Triangle Data" ) )
	{ }

	/***** set all_perf *****/

	else if ( ret = grok_line( str,
			ALL_PERF_PREFIX, &all_perf,
			0, 1, "Set All Performance" ) )
	{ }

	/***** set timing_total *****/

	else if ( ret = grok_line( str,
			TIMING_TOTAL_PREFIX, &timing_total,
			0, 1, "Set Timing Totals" ) )
	{ }

	/***** set timing_average *****/

	else if ( ret = grok_line( str,
			TIMING_AVERAGE_PREFIX, &timing_average,
			0, 1, "Set Timing Averages" ) )
	{ }

	/* return */

	return( ret );
}


int
grok_line( line, prefix, var, min, max, desc )
char	*line;
char	*prefix;
int		*var;
int		min;
int		max;
char	*desc;
{
	char	temp_buf[MAX_LINE_LENGTH];

	int		tmp;
	int		fc;
	int		lc;
	int		i;

	if ( strncmp( line, prefix, strlen( prefix ) ) == 0 )
	{
		fc = strspn( line + strlen( prefix ), " \t:=" )
				+ strlen( prefix );

		lc = strcspn( line + fc, " \t\n" );

		*( line + fc + lc ) = (char) NULL;

		if ( sscanf( line + fc, "%d", &tmp ) == 1 )
			*var = tmp;

		else
		{
			fprintf( stderr, "Error parsing option value\n" );
			return( FAILED );
		}

		if ( max >= 0 && *var > max )
		{
			fprintf( stderr,
					"\t%s Out of Range (%d), using %d by default\n",
					desc, *var, max );
			*var = max;
		}

		if ( min >= 0 && *var < min )
		{
			fprintf( stderr,
					"\t%s Out of Range (%d), using %d by default\n",
					desc, *var, min );
			*var = min;
		}

		return( PASSED );
	}

	return( 0 );
}


/***********************/
/*  Group Server init  */
/***********************/

int
start_group_server()
{
	int	info;

	/* Start pvm group server */

	fprintf( outfile, "Starting pvm group server...  " );
	fflush( outfile );

	info = pvm_joingroup( "Group_init" );

	CKERR_FP( outfile, info, "pvm_joingroup", NOHOST,
			return( FAILED ) );

	fprintf( outfile, "done.\n" );

	return( PASSED );
}


/*************************/
/*  PVM Check host list  */
/*************************/

int
check_hosts()
{
	int		*infos;
	int		info;
	int		i;

	fprintf( outfile, "Building Virtual machine...  " );
	fflush( outfile );

	if ( num_hosts == 1 )
	{
		fprintf( outfile, "done.\n" );

		return( PASSED );
	}

	assert( (infos = (int *) malloc( sizeof( int )
			* ( num_hosts - 1 ) )) != NULL );

	info = pvm_addhosts( hostpool + 1, num_hosts - 1, infos );

	CKERR_FP( outfile, info, "pvm_addhosts", NOHOST,
			{ free( infos ); return( FAILED ); } );

	if ( info < num_hosts - 1 )
	{
		for ( i=0 ; i < num_hosts - 1 ; i++ )
		{
			CKERR_FP( outfile, infos[i], "pvm_addhosts",
					hostpool[ i + 1 ],
					{ free( infos ); return( FAILED ); } );
		}
	}

	free( infos );

	if ( wait_number_hosts( num_hosts ) < 0 )
		return( FAILED );

	fprintf( outfile, "done.\n" );

	/* At this point, all hosts are added */

	return( 0 );
}


/***************************/
/*  Get arch for hostpool  */
/***************************/

int
get_all_arch()
{
	struct pvmhostinfo	*hostp;
	int		nhost;
	int		info;
	int		narch;
	int		i;

	/* Remove all host but master pvmd's host */

	fprintf( outfile, "Getting the ARCH...  " );
	fflush( outfile );

	info = pvm_config( &nhost, &narch, &hostp );

	CKERR_FP( outfile, info, "pvm_config", NOHOST, return( FAILED ) );

	for( i=1 ; i < nhost ; i++ )
	{
		if ( strcmp( hostp[i].hi_name, hostpool[i] ) )
		{
			fprintf( outfile,
		"\tpvm_config(): Response names different from previous\n" );
			return( FAILED );
		}

		hostarch[i] = copy_str( hostp[i].hi_arch );
	}

	fprintf( outfile, "done.\n" );

	return( 0 );
}


/*****************************/
/*  Get number of diff arch  */
/*****************************/

int
get_num_diff_arch()
{
	char	temp[MAX_HOST][25];

	int		result=0;
	int		trouve;
	int		i, j;

	for ( i=1 ; i < num_hosts ; i++ )
	{
		trouve = 0;

		for ( j=0 ; j < result ; j++ )
			trouve = trouve || !strcmp( hostarch[i], temp[j] );

		if ( !trouve )
			strcpy( temp[result++], hostarch[i] );
	}

	for ( i=0 ; i < result ; i++ )
		diff_arch[i] = copy_str( temp[i] );

	return( result );
}


/*****************************/
/*  Check for Multi-Arch VM  */
/*****************************/

int
check_multi_arch()
{
	struct pvmhostinfo *hostp;

	int		narch;
	int		nhost;
	int		info;

	info = pvm_config( &nhost, &narch, &hostp );

	CKERR( info, "pvm_config", NOHOST, return( 1 ) );

	return( ( narch > 1 ) ? 1 : 0 );
}


/**************************/
/*  PVM Remove all tasks  */
/**************************/

int
remove_all_tasks()
{
	struct pvmtaskinfo *taskp;

	int		*noresets = (int *) NULL;
	int		nnr = 0;

	int		count = 0;
	int		tcount;
	int		found;
	int		ntask;
	int		info;
	int		i, j;

#ifdef DEBUG
	fprintf( outfile, "Removing all tasks...\n" );
#endif

	do
	{
		if ( count )
			sleep( 1 );

		info = pvm_tasks( 0, &ntask, &taskp );

		CKERR_FP( outfile, info, "pvm_tasks", NOHOST,
				return( FAILED ) );

		pvm_getnoresets( &noresets, &nnr );

		tcount = ntask;

		for ( i=0 ; i < ntask ; i++ )
		{
			if ( taskp[i].ti_tid != pvm_mytid() )
			{
				for ( j=0, found=0 ; j < nnr && !found ; j++ )
					if ( noresets[j] == taskp[i].ti_tid )
						found++;

				if ( !found )
				{
#ifdef DEBUG
					fprintf( outfile,
							"Killing Task: %s, Tid: 0x%x...\n",
							taskp[i].ti_a_out, taskp[i].ti_tid );
#endif
					info = pvm_kill( taskp[i].ti_tid );

					CKERR_FP( outfile, info, "pvm_kill", NOHOST,
							return( FAILED ) );
				}

				else
					tcount--;
			}
		}
	}
	while ( tcount != 1 && ++count < TIMEOUT_KILL );

	if ( count >= TIMEOUT_KILL )
	{
		fprintf( stderr,
				"Error Removing All Tasks:  Timeout expired.\n" );

		return( FAILED );
	}

	return( 0 );
}


/****************************/
/*  PVM Remove slave tasks  */
/****************************/

int
remove_slave_tasks()
{
	struct pvmtaskinfo *taskp;

	int		*noresets = (int *) NULL;
	int		nnr = 0;

	int		count = 0;
	int		tcount;
	int		found;
	int		ntask;
	int		info;
	int		i, j;

#ifdef DEBUG
	fprintf( outfile, "Removing slave tasks...\n" );
#endif

	do
	{
		if ( count )
			sleep( 1 );

		info = pvm_tasks( 0, &ntask, &taskp );

		CKERR_FP( outfile, info, "pvm_tasks", NOHOST,
				return( FAILED ) );

		pvm_getnoresets( &noresets, &nnr );

		tcount = ntask;

		for ( i=0 ; i < ntask ; i++ )
		{
			if ( taskp[i].ti_tid != pvm_mytid()
					&& strcmp( taskp[i].ti_a_out, "pvmgs" ) )
			{
				for ( j=0, found=0 ; j < nnr && !found ; j++ )
					if ( noresets[j] == taskp[i].ti_tid )
						found++;

				if ( !found )
				{
#ifdef DEBUG
					fprintf( outfile,
							"Killing Task: %s, Tid: 0x%x...\n",
							taskp[i].ti_a_out, taskp[i].ti_tid );
#endif
					info = pvm_kill( taskp[i].ti_tid );

					CKERR_FP( outfile, info, "pvm_kill", NOHOST,
							return( FAILED ) );
				}

				else
					tcount--;
			}
		}
	}
	while ( tcount > 2 && ++count < TIMEOUT_KILL );

	if ( count >= TIMEOUT_KILL )
	{
		fprintf( stderr,
				"Error Removing Slave Tasks:  Timeout expired.\n" );

		return( FAILED );
	}

	return( 0 );
}


/**************************/
/*  PVM Remove all hosts  */
/**************************/

int
remove_all_hosts()
{
	struct pvmhostinfo	*hostp;

	int		narch;
	int		nhost;
	int		info;
	int		i;

	/* Remove all host but master pvmd's host */

#ifdef DEBUG
	fprintf( outfile, "Removing all hosts...\n" );
#endif

	info = pvm_config( &nhost, &narch, &hostp );

	CKERR_FP( outfile, info, "pvm_config", NOHOST, return( FAILED ) );

	for ( i=1 ; i < nhost ; i++ )
	{
#ifdef DEBUG
		fprintf( outfile, "Removing Host: %s, Tid: 0x%x...\n",
				hostp[i].hi_name, hostp[i].hi_tid );
#endif
		info = pvm_delhosts( &(hostp[i].hi_name), 1, &info );

		CKERR_FP( outfile, info, "pvm_delhosts", hostp[i].hi_name,
				return( FAILED ) );
	}

	if ( wait_number_hosts( 1 ) < 0 )
		return( FAILED );

	return( 0 );
}


/*******************/
/*  Check version  */
/*******************/

int
check_version( myversion )
char	*myversion;
{
	char **version_table;
	char *args[2];

	int error_test=0;
	int bufid;
	int info;
	int tid;
	int i;

	fprintf( outfile, "Checking the version...  " );
	fflush( outfile );

	args[0] = test_str( VERSION_TEST );
	args[1] = ( char * ) NULL;

	assert( ( version_table = ( char ** ) malloc( num_hosts
			* sizeof( char * ) ) ) != NULL );

	memset( version_table, 0x00, num_hosts * sizeof( char * ) );

	for ( i=1 ; i < num_hosts ; i++ )
	{
		assert( ( version_table[i] = ( char * ) malloc( 25
				* sizeof( char ) ) ) != NULL );

		memset( version_table[i], 0x00, 25 * sizeof( char ) );
	}

	/* Spawn version_slave */

	for ( i=1 ; i < num_hosts ; i++ )
	{
#ifdef DEBUG
		fprintf( outfile, "Spawning on host %s\n", hostpool[i] );
#endif
		info = pvm_spawn( TEST_SLAVE, args, PvmTaskHost, hostpool[i],
				1, &tid );

		CKERR_FP( outfile, info, "pvm_spawn", hostpool[i],
				{ error_test = FAILED; continue; } );

		info = timeout_wait_message( -1, VERSION_TAG, i, VERBOSE );

		if ( info < 0 )
		{
			error_test = FAILED;
			continue;
		}

		bufid = pvm_recv( -1, VERSION_TAG );

		CKERR_FP( outfile, bufid, "pvm_recv", NOHOST,
				{ error_test = FAILED; continue; } );

		info = pvm_upkstr( version_table[i] );

		CKERR_FP( outfile, info, "pvm_upkstr", NOHOST,
				{ error_test = FAILED; continue; } );

#ifdef DEBUG
		fprintf( outfile, "Host: %s, Version: %s\n",
				hostpool[i], version_table[i] );
#endif
	}

	free( args[0] );

	/* Now the table is complete */

	if ( error_test != FAILED )
	{
		for ( i=1 ; i < num_hosts ; i++ )
		{
			if ( strcmp( myversion, version_table[i] ) != 0 )
			{
				fprintf( outfile,
						"\tVersion mismatch on %s, %s vs. %s\n",
						hostpool[i], version_table[i], myversion );
				error_test = FAILED;
			}
		}
	}

	for ( i=1 ; i < num_hosts ; i++ )
		free( version_table[i] );

	free( version_table );

	fprintf( outfile, "done.\n" );

	return( error_test );
}


/****************************/
/*  Miscellenaous Printout  */
/****************************/

void print_out_misc( myversion )
char	*myversion;
{
	int	i;

	fprintf( outfile, "PVM %s\n", myversion );
	fprintf( outfile, "------------------------------------\n" );

	fprintf( outfile, "host pool\n" );

	fprintf( outfile, "---------\n" );

	fprintf( outfile, "(%s) \t%s \t<--- Master host\n",
			hostarch[0], hostpool[0] );

	for ( i=1 ;i < num_hosts ; i++ )
		fprintf( outfile, "(%s) \t%s\n", hostarch[i], hostpool[i] );

	if ( infilename != NULL )
		fprintf( outfile, "\ninput file:  %s\n", infilename );

	fprintf( outfile, "\noutput file:  %s\n", outfilename );

	fprintf( outfile, "------------------------------------\n" );
}


/******************/
/*  Find Shuffle  */
/******************/

int
find_shuffle( num_diff_arch, shuffle )
int	num_diff_arch;
int	shuffle[ MAX_HOST ];
{
	int		num_computed = 1;
	int		info;
	int		i;
	int		k;

	for ( k=0 ; k < num_hosts ; k++ )
		shuffle[k] = 0;

	while ( num_computed < num_hosts )
	{
		for ( i=0 ; i < num_diff_arch ; i++ )
		{
			info = look_for_new_arch( i, shuffle, num_computed );

			if ( info >= 0 )
				shuffle[ num_computed++ ] = info;
		}
	}

	return( PASSED );
}


int	look_for_new_arch( arch, shuffle, num_computed )
int	arch;
int	shuffle[ MAX_HOST ];
int	num_computed;
{
	int		test = 0;
	int		i, j;

	for ( i=1 ; i < num_hosts ; i++ )
	{
		test = 0;

		if ( !strcmp( hostarch[i], diff_arch[arch] ) )
		{
			for( j=1 ; j < num_computed ; j++ )
			{
				if ( shuffle[j] == i )
					test = 1;
			}

			if ( test == 0 )
				return( i );
		}
	}

	return( FAILED );
}


/*************************/
/*  Allocate some hosts  */
/*************************/

int
allocate( array, req, num_wanted )
int	array[MAX_HOST];
int	req;
int	*num_wanted;
{
	int		error_test=0;
	int		count_zero=0;
	int		infos;
	int		info;
	int		i;

#ifdef DEBUG
	errprintf( "allocate(): requested %d, wanted after %d\n",
			req, *num_wanted );
#endif

	for ( i=1 ; i < req ; i++ )
	{
		info = pvm_addhosts( &(hostpool[ array[i] ]), 1, &infos );

		CKERR_FP( outfile, info, "pvm_addhosts", hostpool[ array[i] ],
				error_test = FAILED );
	}

	if ( error_test == FAILED )
		return( FAILED );

	if ( wait_number_hosts( req ) < 0 )
		return( FAILED );

	return( error_test );
}


/****************************/
/*  Unexpected host killed  */
/****************************/

int
unexpected_host_killed( num_wanted )
int	num_wanted;
{
	struct	pvmhostinfo *hostp;

	int		after;
	int		narch;
	int		info;

	info = pvm_config( &after, &narch, &hostp );

	CKERR_FP( outfile, info, "pvm_config", NOHOST, return( -2 ) );

	/* We suppress the master */

	after--;

	if ( after != num_wanted )
	{
		errprintf( "\t%d hosts expected, %d found\n",
				num_wanted, after );
		return( FAILED );
	}

	return( PASSED );
}


/***********************/
/*  Print test result  */
/***********************/

void
print_test_result( test_num )
int		test_num;
{
	int status;

	status = module_list[ test_num ].status;

	switch ( status )
	{
		case NOT_YET_PROCESSED:
			fprintf( outfile, "-> NOT_YET_PROCESSED\n" );
			break;

		case PASSED:
			fprintf( outfile, "-> PASSED\n" );
			break;

		case FAILED:
			fprintf( outfile, "-> FAILED\n" );
			break;

		case NOT_COMPLETED:
			fprintf( outfile, "-> NOT_COMPLETED\n" );
			break;

		case NOT_POSSIBLE:
			fprintf( outfile, "-> NOT_POSSIBLE\n" );
			break;

		default:
			fprintf( outfile, "-> Undefined Return Status = %d\n",
					status );
			break;
	}
}


/********************************/
/*  Wait for a number of hosts  */
/********************************/

int
wait_number_hosts( number )
int	number;
{
	struct	pvmhostinfo *hostp;

	int		actual_num;
	int		count = 0;
	int		narch;
	int		info;

	info = pvm_config( &actual_num, &narch, &hostp );

	CKERR_FP( outfile, info, "pvm_config", NOHOST, return( FAILED ) );

#ifdef DEBUG
	errprintf( "wait_number_hosts(): Wanted %d, got %d.\n",
			number, actual_num );
#endif

	while( count <= TIMEOUT_ADD_DELETE && actual_num != number )
	{
		info = pvm_config( &actual_num, &narch, &hostp );

		CKERR_FP( outfile, info, "pvm_config", NOHOST,
				return( FAILED ) );

#ifdef DEBUG
		errprintf( "wait_number_hosts(): Wanted %d, got %d.\n\n",
				number, actual_num );
#endif

		sleep( 1 );

		count++;
	}

	if ( actual_num != number )
	{
		errprintf( "\tAdd/Del Timeout exceeded.\n" );
		return( -2 );
	}

	return( PASSED );
}


/**************************/
/*  timeout_wait_message  */
/**************************/

int
timeout_wait_message( tid, tag, host_num, verbose )
int		tid;
int		tag;
int		host_num;
int		verbose;
{
	int		count = 0;
	int		info;

	while ( 1 )
	{
		info = pvm_probe( tid, tag );

		CKERR_FP( outfile, info, "pvm_probe", NOHOST,
				return( FAILED ) );

		if ( info > 0 )
			return( PASSED );

		else if ( info == 0 )
		{
			if ( count > TIMEOUT_MESSAGE )
			{
				if ( verbose == VERBOSE )
				{
					if ( host_num == -1 )
						errprintf( "\tMessage Timeout expired\n" );

					else
					{
						errprintf( 
							"\tMessage Timeout expired from host %s\n",
							hostpool[ host_num ] );
					}
				}

				return( -2 );
			}

			count++;

			sleep( 1 );
		}
	}
}


/***********************/
/*  timeout_wait_kill  */
/***********************/

int
timeout_wait_kill( tid, host_num )
int		tid;
int		host_num;
{
	int		count=0;
	int		info;

	while ( 1 )
	{
		info = pvm_pstat( tid );

		if ( info < 0 ) break;

		if ( count > TIMEOUT_KILL ) break;

		sleep( 1 );

		count++;
	}

	if ( info != PvmNoTask && count < TIMEOUT_KILL )
	{
		if ( host_num == -1 )
		{
			CKERR_FP( outfile, info, "pvm_pstat", NOHOST,
					return( FAILED ) );
		}

		else
		{
			CKERR_FP( outfile, info, "pvm_pstat", hostpool[ host_num ],
					return( FAILED ) );
		}
	}

	if ( count >= TIMEOUT_KILL )
	{
		if ( host_num == -1 )
			errprintf( "\tKill Timeout expired\n" );

		else
		{
			errprintf( "\tKill Timeout expired for host %s\n",
					hostpool[ host_num ] );
		}

		return( -2 );
	}

	if ( info != PvmNoTask )
	{
		CKERR_FP( outfile, info, "pvm_pstat", NOHOST,
				return( FAILED ) );
	}

	return( PASSED );
}


/******************/
/*  Test Cleanup  */
/******************/

int
test_cleanup( test_num )
int test_num;
{
	int		msgtag;
	int		nbytes;
	int		bufid;
	int		info;
	int		ret;
	int		src;

	ret = 0;

	if ( remove_all_hosts() < 0 )
	{
		errprintf( "\tCannot clean up (delete) hosts\n" );

		ret = FAILED;
	}

	if ( remove_slave_tasks() < 0 )
	{
		errprintf( "\tCannot clean up (kill) slave tasks\n" );

		ret = FAILED;
	}

	while ( (info = pvm_probe( -1, -1 )) > 0 )
	{
		bufid = pvm_recv( -1, -1 );
		CKERR( bufid, "pvm_recv", NOHOST, return( FAILED ) );

		info = pvm_bufinfo( bufid, &nbytes, &msgtag, &src );
		CKERR( info, "pvm_bufinfo", NOHOST, return( FAILED ) );

		errprintf( "\tWarning:  Leftover Messages After Test " );
		errprintf( "(src=0x%x msgtag=%d nbytes=%d)\n",
				src, msgtag, nbytes );

		ret = FAILED;
	}

	CKERR( info, "pvm_probe", NOHOST, ret = FAILED );

	if ( ret < 0 )
	{
		errprintf( "\tError Cleaning Up After Test #%d (%s)!!!\n",
				test_num + 1,
				module_list[ test_num ].description );
	}

	return( ret );
}


/***************************/
/*  Buffered Error-Printf  */
/***************************/

int
#ifdef __STDC__
errprintf( const char *fmt, ... )
#else
errprintf( va_alist )
	va_dcl
#endif
{
	va_list ap;

	test_errmsg *EMptr;
	test_errmsg *EM;

	int		cc;

#ifdef __STDC__

	va_start( ap, fmt );

#else

	char	*fmt;

	va_start( ap );

	fmt = va_arg( ap, char * );

#endif

	EM = (test_errmsg *) malloc( sizeof( test_errmsg ) );

	if ( !memcheck( EM, "Test Error Message Storage" ) )
		return( 0 );

	cc = vsprintf( EM->errmsg, fmt, ap );

	EM->next = ( test_errmsg * ) NULL;

	if ( ERRMSGS != NULL )
	{
		EMptr = ERRMSGS;

		while ( EMptr->next != NULL )
			EMptr = EMptr->next;

		EMptr->next = EM;
	}

	else
		ERRMSGS = EM;

	va_end( ap );

	return cc;
}


void
dump_errmsgs()
{
	test_errmsg *EMnext;
	test_errmsg *EM;

	EM = ERRMSGS;

	ERRMSGS = ( test_errmsg * ) NULL;

	while ( EM != NULL )
	{
		fprintf( outfile, "%s", EM->errmsg );

		EMnext = EM->next;
		free( EM );
		EM = EMnext;
	}

	fflush( outfile );
}


/*--------------------------------------*/
/*--------- Internal Stuff -------------*/
/*--------------------------------------*/

/* parse_flag */
int
parse_flag( line )
char	*line;
{
	char	*lineptr;

	int		fc;

	lineptr = line + 1;

	fc = strspn( lineptr, " \t" );

	if ( lineptr[ fc ] == ']' )
		return( NOPERFORM );

	else
	{
		lineptr++;

		if ( strcspn( lineptr, "]" ) == strlen( lineptr ) )
			return( FAILED );

		else
			return( PERFORM );
	}
}


/* parse_test_pattern */
char *
parse_test_pattern( line )
char	*line;
{
	char	tmp[1024];

	char	*lineptr;
	char	*retstr;

	char	c;

	int		test_num;
	int		len;
	int		fc;

	fc = strcspn( line, "]" );

	lineptr = line + fc + 1;

	fc = strspn( lineptr, " \t" );

	lineptr += fc;

	len = strlen( lineptr );

	c = lineptr[ len - 1 ];

	lineptr[ len - 1 ] = '\0';

#ifdef USE_GNU_REGEX
	sprintf( tmp, ".*%s.*", lineptr );
#else
	sprintf( tmp, "*%s*", lineptr );
#endif

	retstr = copy_str( tmp );

	lineptr[ len - 1 ] = c;

	return( retstr );
}


char *
test_str( test_num )
int		test_num;
{
	char	tmp[32];

	sprintf( tmp, "%d", test_num );

	return( copy_str( tmp ) );
}


char *
copy_str( str )
char	*str;
{
	char	*tmp;

	tmp = ( char * ) malloc( ( unsigned ) ( strlen( str ) + 1 )
		* sizeof( char ) );

	if ( !memcheck( tmp, "Copy String" ) )
		return( (char *) NULL );

	strcpy( tmp, str );

	return( tmp );
}


int
filecheck( fp, name )
FILE	*fp;
char	*name;
{
	if ( fp == NULL )
	{
		fprintf( stderr, "\nError Opening File \"%s\"\n\n", name );

		return( 0 );
	}

	return( 1 );
}


int
memcheck( ptr, name )
void	*ptr;
char	*name;
{
	if ( ptr == NULL )
	{
		fprintf( stderr, "\nError Allocating Memory for \"%s\"\n\n",
			name );

		return( 0 );
	}

	return( 1 );
}

